# Copyright 2017, 2018, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from digi.xbee.packets.base import XBeeAPIPacket, DictKeys
from digi.xbee.models.address import XBee64BitAddress, XBee16BitAddress
from digi.xbee.models.status import TransmitStatus
from digi.xbee.exception import InvalidOperatingModeException, InvalidPacketException
from digi.xbee.packets.aft import ApiFrameType
from digi.xbee.models.mode import OperatingMode
from digi.xbee.io import IOSample, IOLine
from digi.xbee.util import utils


class TX64Packet(XBeeAPIPacket):
    """
    This class represents a TX (Transmit) 64 Request packet. Packet is built 
    using the parameters of the constructor or providing a valid byte array.
    
    A TX Request message will cause the module to transmit data as an RF 
    Packet.
    
    .. seealso::
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 15

    def __init__(self, frame_id, x64bit_addr, transmit_options, rf_data):
        """
        Class constructor. Instantiates a new :class:`.TX64Packet` object with the provided parameters.

        Args:
            frame_id (Integer): the frame ID of the packet.
            x64bit_addr (:class:`.XBee64BitAddress`): the 64-bit destination address.
            transmit_options (Integer): bitfield of supported transmission options.
            rf_data (Bytearray, optional): RF data that is sent to the destination device. Optional.

        .. seealso::
           | :class:`.TransmitOptions`
           | :class:`.XBee64BitAddress`
           | :class:`.XBeeAPIPacket`

        Raises:
            ValueError: if ``frame_id`` is less than 0 or greater than 255.
        """
        if frame_id < 0 or frame_id > 255:
            raise ValueError("Frame id must be between 0 and 255.")

        super().__init__(ApiFrameType.TX_64)
        self._frame_id = frame_id
        self.__x64bit_addr = x64bit_addr
        self.__transmit_options = transmit_options
        self.__rf_data = rf_data

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.
        
        Returns:
            :class:`.TX64Packet`.
            
        Raises:
            InvalidPacketException: if the bytearray length is less than 15. (start delim. + length (2 bytes) + frame
                type + frame id + 64bit addr. + transmit options + checksum = 15 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :attr:`.ApiFrameType.TX_64`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.
            
        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=TX64Packet.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.TX_64.code:
            raise InvalidPacketException("This packet is not a TX 64 packet.")

        return TX64Packet(raw[4], XBee64BitAddress(raw[5:13]), raw[13], raw[14:-1])

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return True

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        ret = self.__x64bit_addr.address
        ret.append(self.__transmit_options)
        if self.__rf_data is not None:
            ret += self.__rf_data
        return ret

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        return {DictKeys.X64BIT_ADDR:      self.__x64bit_addr.address,
                DictKeys.TRANSMIT_OPTIONS: self.__transmit_options,
                DictKeys.RF_DATA:          self.__rf_data}

    def __get_64bit_addr(self):
        """
        Returns the 64-bit destination address.
        
        Returns:
            :class:`.XBee64BitAddress`: the 64-bit destination address.

        .. seealso::
           | :class:`.XBee64BitAddress`
        """
        return self.__x64bit_addr

    def __set_64bit_addr(self, x64bit_addr):
        """
        Sets the 64-bit destination address.

        Args:
            x64bit_addr (:class:`.XBee64BitAddress`): the new 64-bit destination address.

        .. seealso::
           | :class:`.XBee64BitAddress`
        """
        self.__x64bit_addr = x64bit_addr

    def __get_transmit_options(self):
        """
        Returns the transmit options bitfield.

        Returns:
            Integer: the transmit options bitfield.

        .. seealso::
           | :class:`.TransmitOptions`
        """
        return self.__transmit_options

    def __set_transmit_options(self, transmit_options):
        """
        Sets the transmit options bitfield.

        Args:
            transmit_options (Integer): the new transmit options bitfield.

        .. seealso::
           | :class:`.TransmitOptions`
        """
        self.__transmit_options = transmit_options

    def __get_rf_data(self):
        """
        Returns the RF data to send.

        Returns:
            Bytearray: the RF data to send.
        """
        if self.__rf_data is None:
            return None
        return self.__rf_data.copy()

    def __set_rf_data(self, rf_data):
        """
        Sets the RF data to send.

        Args:
            rf_data (Bytearray): the new RF data to send.
        """
        if rf_data is None:
            self.__rf_data = None
        else:
            self.__rf_data = rf_data.copy()

    x64bit_dest_addr = property(__get_64bit_addr, __set_64bit_addr)
    """XBee64BitAddress. 64-bit destination address."""

    transmit_options = property(__get_transmit_options, __set_transmit_options)
    """Integer. Transmit options bitfield."""

    rf_data = property(__get_rf_data, __set_rf_data)
    """Bytearray. RF data to send."""


class TX16Packet(XBeeAPIPacket):
    """
    This class represents a TX (Transmit) 16 Request packet. Packet is built 
    using the parameters of the constructor or providing a valid byte array.
    
    A TX request message will cause the module to transmit data as an RF
    packet.
    
    .. seealso::
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 9

    def __init__(self, frame_id, x16bit_addr, transmit_options, rf_data=None):
        """
        Class constructor. Instantiates a new :class:`.TX16Packet` object with the provided parameters.

        Args:
            frame_id (Integer): the frame ID of the packet.
            x16bit_addr (:class:`.XBee16BitAddress`): the 16-bit destination address.
            transmit_options (Integer): bitfield of supported transmission options.
            rf_data (Bytearray, optional): RF data that is sent to the destination device. Optional.

        .. seealso::
           | :class:`.TransmitOptions`
           | :class:`.XBee16BitAddress`
           | :class:`.XBeeAPIPacket`

        Raises:
            ValueError: if ``frame_id`` is less than 0 or greater than 255.
        """
        if frame_id < 0 or frame_id > 255:
            raise ValueError("Frame id must be between 0 and 255.")

        super().__init__(ApiFrameType.TX_16)
        self._frame_id = frame_id
        self.__x16bit_addr = x16bit_addr
        self.__transmit_options = transmit_options
        self.__rf_data = rf_data

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.
        
        Returns:
            :class:`.TX16Packet`.
            
        Raises:
            InvalidPacketException: if the bytearray length is less than 9. (start delim. + length (2 bytes) + frame
                type + frame id + 16bit addr. + transmit options + checksum = 9 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :attr:`.ApiFrameType.TX_16`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.
            
        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=TX16Packet.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.TX_16.code:
            raise InvalidPacketException("This packet is not a TX 16 packet.")

        return TX16Packet(raw[4], XBee16BitAddress(raw[5:7]), raw[7], raw[8:-1])

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return True

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        ret = self.__x16bit_addr.address
        ret.append(self.__transmit_options)
        if self.__rf_data is not None:
            ret += self.__rf_data
        return ret

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        return {DictKeys.X16BIT_ADDR:      self.__x16bit_addr,
                DictKeys.TRANSMIT_OPTIONS: self.__transmit_options,
                DictKeys.RF_DATA:          self.__rf_data}

    def __get_16bit_addr(self):
        """
        Returns the 16-bit destination address.

        Returns:
            :class:`.XBee16BitAddress`: the 16-bit destination address.

        .. seealso::
           | :class:`.XBee16BitAddress`
        """
        return self.__x16bit_addr

    def __set_16bit_addr(self, x16bit_addr):
        """
        Sets the 16-bit destination address.

        Args:
            x16bit_addr (:class:`.XBee16BitAddress`): the new 16-bit destination address.

        .. seealso::
           | :class:`.XBee16BitAddress`
        """
        self.__x16bit_addr = x16bit_addr

    def __get_transmit_options(self):
        """
        Returns the transmit options bitfield.

        Returns:
            Integer: the transmit options bitfield.

        .. seealso::
           | :class:`.TransmitOptions`
        """
        return self.__transmit_options

    def __set_transmit_options(self, transmit_options):
        """
        Sets the transmit options bitfield.

        Args:
            transmit_options (Integer): the new transmit options bitfield.

        .. seealso::
           | :class:`.TransmitOptions`
        """
        self.__transmit_options = transmit_options

    def __get_rf_data(self):
        """
        Returns the RF data to send.

        Returns:
            Bytearray: the RF data to send.
        """
        if self.__rf_data is None:
            return None
        return self.__rf_data.copy()

    def __set_rf_data(self, rf_data):
        """
        Sets the RF data to send.

        Args:
            rf_data (Bytearray): the new RF data to send.
        """
        if rf_data is None:
            self.__rf_data = None
        else:
            self.__rf_data = rf_data.copy()

    x16bit_dest_addr = property(__get_16bit_addr, __set_16bit_addr)
    """XBee64BitAddress. 16-bit destination address."""

    transmit_options = property(__get_transmit_options, __set_transmit_options)
    """Integer. Transmit options bitfield."""

    rf_data = property(__get_rf_data, __set_rf_data)
    """Bytearray. RF data to send."""


class TXStatusPacket(XBeeAPIPacket):
    """
    This class represents a TX (Transmit) status packet. Packet is built using
    the parameters of the constructor or providing a valid API payload.

    When a TX request is completed, the module sends a TX status message.
    This message will indicate if the packet was transmitted successfully or if
    there was a failure.

    .. seealso::
       | :class:`.TX16Packet`
       | :class:`.TX64Packet`
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 7

    def __init__(self, frame_id, transmit_status):
        """
        Class constructor. Instantiates a new :class:`.TXStatusPacket` object with the provided parameters.

        Args:
            frame_id (Integer): the frame ID of the packet.
            transmit_status (:class:`.TransmitStatus`): transmit status. Default: SUCCESS.

        Raises:
            ValueError: if ``frame_id`` is less than 0 or greater than 255.

        .. seealso::
           | :class:`.TransmitStatus`
           | :class:`.XBeeAPIPacket`
        """
        if frame_id < 0 or frame_id > 255:
            raise ValueError("Frame id must be between 0 and 255.")

        super().__init__(ApiFrameType.TX_STATUS)
        self._frame_id = frame_id
        self.__transmit_status = transmit_status

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.

        Returns:
            :class:`.TXStatusPacket`.

        Raises:
            InvalidPacketException: if the bytearray length is less than 7. (start delim. + length (2 bytes) + frame
                type + frame id + transmit status + checksum = 7 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :attr:`.ApiFrameType.TX_16`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.

        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=TXStatusPacket.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.TX_STATUS.code:
            raise InvalidPacketException("This packet is not a TX status packet.")

        return TXStatusPacket(raw[4], TransmitStatus.get(raw[5]))

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return True

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        return utils.int_to_bytes(self.__transmit_status.code, 1)

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        return {DictKeys.TS_STATUS: self.__transmit_status}

    def __get_transmit_status(self):
        """
        Returns the transmit status.

        Returns:
            :class:`.TransmitStatus`: the transmit status.

        .. seealso::
           | :class:`.TransmitStatus`
        """
        return self.__transmit_status

    def __set_transmit_status(self, transmit_status):
        """
        Sets the transmit status.

        Args:
            transmit_status (:class:`.TransmitStatus`): the new transmit status to set.

        .. seealso::
           | :class:`.TransmitStatus`
        """
        self.__transmit_status = transmit_status

    transmit_status = property(__get_transmit_status, __set_transmit_status)
    """:class:`.TransmitStatus`. Transmit status."""


class RX64Packet(XBeeAPIPacket):
    """
    This class represents an RX (Receive) 64 request packet. Packet is built
    using the parameters of the constructor or providing a valid API byte array.
    
    When the module receives an RF packet, it is sent out the UART using 
    this message type.
    
    This packet is the response to TX (transmit) 64 request packets.
    
    .. seealso::
       | :class:`.ReceiveOptions`
       | :class:`.TX64Packet`
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 15

    def __init__(self, x64bit_addr, rssi, receive_options, rf_data=None):
        """
        Class constructor. Instantiates a :class:`.RX64Packet` object with the provided parameters.
        
        Args:
            x64bit_addr (:class:`.XBee64BitAddress`): the 64-bit source address.
            rssi (Integer): received signal strength indicator.
            receive_options (Integer): bitfield indicating the receive options.
            rf_data (Bytearray, optional): received RF data. Optional.

        .. seealso::
           | :class:`.ReceiveOptions`
           | :class:`.XBee64BitAddress`
           | :class:`.XBeeAPIPacket`
        """

        super().__init__(ApiFrameType.RX_64)

        self.__x64bit_addr = x64bit_addr
        self.__rssi = rssi
        self.__receive_options = receive_options
        self.__rf_data = rf_data

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.
        
        Returns:
            :class:`.RX64Packet`
            
        Raises:
            InvalidPacketException: if the bytearray length is less than 15. (start delim. + length (2 bytes) + frame
                type + 64bit addr. + rssi + receive options + checksum = 15 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :attr:`.ApiFrameType.RX_64`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.
            
        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=RX64Packet.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.RX_64.code:
            raise InvalidPacketException("This packet is not an RX 64 packet.")

        return RX64Packet(XBee64BitAddress(raw[4:12]), raw[12], raw[13], raw[14:-1])

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return False

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        ret = self.__x64bit_addr.address
        ret.append(self.__rssi)
        ret.append(self.__receive_options)
        if self.__rf_data is not None:
            ret += self.__rf_data
        return ret

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        return {DictKeys.X64BIT_ADDR:     self.__x64bit_addr,
                DictKeys.RSSI:            self.__rssi,
                DictKeys.RECEIVE_OPTIONS: self.__receive_options,
                DictKeys.RF_DATA:         self.__rf_data}

    def __get_64bit_addr(self):
        """
        Returns the 64-bit source address.

        Returns:
            :class:`.XBee64BitAddress`: the 64-bit source address.

        .. seealso::
           | :class:`.XBee64BitAddress`
        """
        return self.__x64bit_addr

    def __set_64bit_addr(self, x64bit_addr):
        """
        Sets the 64-bit source address.

        Args:
            x64bit_addr (:class:`.XBee64BitAddress`): the new 64-bit source address.

        .. seealso::
           | :class:`.XBee64BitAddress`
        """
        self.__x64bit_addr = x64bit_addr

    def __get_rssi(self):
        """
        Returns the received Signal Strength Indicator (RSSI).
        
        Returns:
            Integer: the received Signal Strength Indicator (RSSI).
        """
        return self.__rssi

    def __set_rssi(self, rssi):
        """
        Sets the received Signal Strength Indicator (RSSI).
        
        Args:
            rssi (Integer): the new received Signal Strength Indicator (RSSI).
        """
        self.__rssi = rssi

    def __get_options(self):
        """
        Returns the receive options bitfield.

        Returns:
            Integer: the receive options bitfield.

        .. seealso::
           | :class:`.ReceiveOptions`
        """
        return self.__receive_options

    def __set_options(self, receive_options):
        """
        Sets the receive options bitfield.

        Args:
            receive_options (Integer): the new receive options bitfield.

        .. seealso::
           | :class:`.ReceiveOptions`
        """
        self.__receive_options = receive_options

    def __get_rf_data(self):
        """
        Returns the received RF data.

        Returns:
            Bytearray: the received RF data.
        """
        if self.__rf_data is None:
            return None
        return self.__rf_data.copy()

    def __set_rf_data(self, rf_data):
        """
        Sets the received RF data.

        Args:
            rf_data (Bytearray): the new received RF data.
        """
        if rf_data is None:
            self.__rf_data = None
        else:
            self.__rf_data = rf_data.copy()

    x64bit_source_addr = property(__get_64bit_addr, __set_64bit_addr)
    """:class:`.XBee64BitAddress`. 64-bit source address."""

    rssi = property(__get_rssi, __set_rssi)
    """Integer. Received Signal Strength Indicator (RSSI) value."""

    receive_options = property(__get_options, __set_options)
    """Integer. Receive options bitfield."""

    rf_data = property(__get_rf_data, __set_rf_data)
    """Bytearray. Received RF data."""


class RX16Packet(XBeeAPIPacket):
    """
    This class represents an RX (Receive) 16 Request packet. Packet is built 
    using the parameters of the constructor or providing a valid API byte array.
    
    When the module receives an RF packet, it is sent out the UART using this 
    message type
    
    This packet is the response to TX (Transmit) 16 Request packets.
    
    .. seealso::
       | :class:`.ReceiveOptions`
       | :class:`.TX16Packet`
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 9

    def __init__(self, x16bit_addr, rssi, receive_options, rf_data=None):
        """
        Class constructor. Instantiates a :class:`.RX16Packet` object with the provided parameters.

        Args:
            x16bit_addr (:class:`.XBee16BitAddress`): the 16-bit source address.
            rssi (Integer): received signal strength indicator.
            receive_options (Integer): bitfield indicating the receive options.
            rf_data (Bytearray, optional): received RF data. Optional.

        .. seealso::
           | :class:`.ReceiveOptions`
           | :class:`.XBee16BitAddress`
           | :class:`.XBeeAPIPacket`
        """

        super().__init__(ApiFrameType.RX_16)

        self.__x16bit_addr = x16bit_addr
        self.__rssi = rssi
        self.__receive_options = receive_options
        self.__rf_data = rf_data

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.
        
        Returns:
            :class:`.RX16Packet`.
            
        Raises:
            InvalidPacketException: if the bytearray length is less than 9. (start delim. + length (2 bytes) + frame
                type + 16bit addr. + rssi + receive options + checksum = 9 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :attr:`.ApiFrameType.RX_16`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.
            
        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=RX16Packet.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.RX_16.code:
            raise InvalidPacketException("This packet is not an RX 16 Packet")

        return RX16Packet(XBee16BitAddress(raw[4:6]), raw[6], raw[7], raw[8:-1])

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return False

    def _get_api_packet_spec_data(self):
        """
        Override method.
        
        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        ret = self.__x16bit_addr.address
        ret.append(self.__rssi)
        ret.append(self.__receive_options)
        if self.__rf_data is not None:
            ret += self.__rf_data
        return ret

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.
        
        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        return {DictKeys.X16BIT_ADDR:     self.__x16bit_addr,
                DictKeys.RSSI:            self.__rssi,
                DictKeys.RECEIVE_OPTIONS: self.__receive_options,
                DictKeys.RF_DATA:         self.__rf_data}

    def __get_16bit_addr(self):
        """
        Returns the 16-bit source address.

        Returns:
            :class:`.XBee16BitAddress`: the 16-bit source address.

        .. seealso::
           | :class:`.XBee16BitAddress`
        """
        return self.__x16bit_addr

    def __set_16bit_addr(self, x16bit_addr):
        """
        Sets the 16-bit source address.

        Args:
            x16bit_addr (:class:`.XBee16BitAddress`): the new 16-bit source address.

        .. seealso::
           | :class:`.XBee16BitAddress`
        """
        self.__x16bit_addr = x16bit_addr

    def __get_rssi(self):
        """
        Returns the received Signal Strength Indicator (RSSI).

        Returns:
            Integer: the received Signal Strength Indicator (RSSI).
        """
        return self.__rssi

    def __set_rssi(self, rssi):
        """
        Sets the received Signal Strength Indicator (RSSI).

        Args:
            rssi (Integer): the new received Signal Strength Indicator (RSSI).

        """
        self.__rssi = rssi

    def __get_options(self):
        """
        Returns the receive options bitfield.

        Returns:
            Integer: the receive options bitfield.

        .. seealso::
           | :class:`.ReceiveOptions`
        """
        return self.__receive_options

    def __set_options(self, receive_options):
        """
        Sets the receive options bitfield.

        Args:
            receive_options (Integer): the new receive options bitfield.

        .. seealso::
           | :class:`.ReceiveOptions`
        """
        self.__receive_options = receive_options

    def __get_rf_data(self):
        """
        Returns the received RF data.

        Returns:
            Bytearray: the received RF data.
        """
        if self.__rf_data is None:
            return None
        return self.__rf_data.copy()

    def __set_rf_data(self, rf_data):
        """
        Sets the received RF data.

        Args:
            rf_data (Bytearray): the new received RF data.
        """
        if rf_data is None:
            self.__rf_data = None
        else:
            self.__rf_data = rf_data.copy()

    x16bit_source_addr = property(__get_16bit_addr, __set_16bit_addr)
    """:class:`.XBee16BitAddress`. 16-bit source address."""

    rssi = property(__get_rssi, __set_rssi)
    """Integer. Received Signal Strength Indicator (RSSI) value."""

    receive_options = property(__get_options, __set_options)
    """Integer. Receive options bitfield."""

    rf_data = property(__get_rf_data, __set_rf_data)
    """Bytearray. Received RF data."""


class RX64IOPacket(XBeeAPIPacket):
    """
    This class represents an RX64 address IO packet. Packet is built using the
    parameters of the constructor or providing a valid API payload.

    I/O data is sent out the UART using an API frame.

    .. seealso::
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 20

    def __init__(self, x64bit_addr, rssi, receive_options, rf_data):
        """
        Class constructor. Instantiates an :class:`.RX64IOPacket` object with the provided parameters.

        Args:
            x64bit_addr (:class:`.XBee64BitAddress`): the 64-bit source address.
            rssi (Integer): received signal strength indicator.
            receive_options (Integer): bitfield indicating the receive options.
            rf_data (Bytearray): received RF data.

        .. seealso::
           | :class:`.ReceiveOptions`
           | :class:`.XBee64BitAddress`
           | :class:`.XBeeAPIPacket`
        """
        super().__init__(ApiFrameType.RX_IO_64)
        self.__x64bit_addr = x64bit_addr
        self.__rssi = rssi
        self.__receive_options = receive_options
        self.__rf_data = rf_data
        self.__io_sample = IOSample(rf_data) if rf_data is not None and len(rf_data) >= 5 else None

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.

        Returns:
            :class:`.RX64IOPacket`.

        Raises:
            InvalidPacketException: if the bytearray length is less than 20. (start delim. + length (2 bytes) + frame
                type + 64bit addr. + rssi + receive options + rf data (5 bytes) + checksum = 20 bytes)
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :attr:`.ApiFrameType.RX_IO_64`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.

        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=RX64IOPacket.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.RX_IO_64.code:
            raise InvalidPacketException("This packet is not an RX 64 IO packet.")

        return RX64IOPacket(XBee64BitAddress(raw[4:12]), raw[12], raw[13], raw[14:-1])

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return False

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        ret = self.__x64bit_addr.address
        ret.append(self.__rssi)
        ret.append(self.__receive_options)
        if self.__rf_data is not None:
            ret += self.__rf_data
        return ret

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        base = {DictKeys.X16BIT_ADDR:         self.__x64bit_addr.address,
                DictKeys.RSSI:                self.__rssi,
                DictKeys.RECEIVE_OPTIONS:     self.__receive_options}

        if self.__io_sample is not None:
            base[DictKeys.NUM_SAMPLES] = 1
            base[DictKeys.DIGITAL_MASK] = self.__io_sample.digital_mask
            base[DictKeys.ANALOG_MASK] = self.__io_sample.analog_mask

            # Digital values
            for i in range(16):
                if self.__io_sample.has_digital_value(IOLine.get(i)):
                    base[IOLine.get(i).description + "digital value"] = \
                        utils.hex_to_string(self.__io_sample.get_digital_value(IOLine.get(i)))

            # Analog values
            for i in range(6):
                if self.__io_sample.has_analog_value(IOLine.get(i)):
                    base[IOLine.get(i).description + "analog value"] = \
                        utils.hex_to_string(self.__io_sample.get_analog_value(IOLine.get(i)))

            # Power supply
            if self.__io_sample.has_power_supply_value():
                base["Power supply value "] = "%02X" % self.__io_sample.power_supply_value

        elif self.__rf_data is not None:
            base[DictKeys.RF_DATA] = utils.hex_to_string(self.__rf_data)

        return base

    def __get_64bit_addr(self):
        """
        Returns the 64-bit source address.

        Returns:
            :class:`XBee64BitAddress`: the 64-bit source address.

        .. seealso::
           | :class:`.XBee64BitAddress`
        """
        return self.__x64bit_addr

    def __set_64bit_addr(self, x64bit_addr):
        """
        Sets the 64-bit source address.

        Args:
            x64bit_addr (:class:`.XBee64BitAddress`): the new 64-bit source address.

        .. seealso::
           | :class:`.XBee64BitAddress`
        """
        self.__x64bit_addr = x64bit_addr

    def __get_rssi(self):
        """
        Returns the received Signal Strength Indicator (RSSI).

        Returns:
            Integer: the received Signal Strength Indicator (RSSI).
        """
        return self.__rssi

    def __set_rssi(self, rssi):
        """
        Sets the received Signal Strength Indicator (RSSI).

        Args:
            rssi (Integer): the new received Signal Strength Indicator (RSSI).
        """
        self.__rssi = rssi

    def __get_options(self):
        """
        Returns the receive options bitfield.

        Returns:
            Integer: the receive options bitfield.

        .. seealso::
           | :class:`.ReceiveOptions`
        """
        return self.__receive_options

    def __set_options(self, receive_options):
        """
        Sets the receive options bitfield.

        Args:
            receive_options (Integer): the new receive options bitfield.

        .. seealso::
           | :class:`.ReceiveOptions`
        """
        self.__receive_options = receive_options

    def __get_rf_data(self):
        """
        Returns the received RF data.

        Returns:
            Bytearray: the received RF data.
        """
        if self.__rf_data is None:
            return None
        return self.__rf_data.copy()

    def __set_rf_data(self, rf_data):
        """
        Sets the received RF data.

        Args:
            rf_data (Bytearray): the new received RF data.
        """
        if rf_data is None:
            self.__rf_data = None
        else:
            self.__rf_data = rf_data.copy()

        # Modify the ioSample accordingly
        if rf_data is not None and len(rf_data) >= 5:
            self.__io_sample = IOSample(self.__rf_data)
        else:
            self.__io_sample = None

    def __get_io_sample(self):
        """
        Returns the IO sample corresponding to the data contained in the packet.

        Returns:
            :class:`.IOSample`: the IO sample of the packet, ``None`` if the packet has not any data or if the
                sample could not be generated correctly.

        .. seealso::
           | :class:`.IOSample`
        """
        return self.__io_sample

    def __set_io_sample(self, io_sample):
        """
        Sets the IO sample of the packet.

        Args:
            io_sample (:class:`.IOSample`): the new IO sample to set.

        .. seealso::
           | :class:`.IOSample`
        """
        self.__io_sample = io_sample

    x64bit_source_addr = property(__get_64bit_addr, __set_64bit_addr)
    """:class:`.XBee64BitAddress`. 64-bit source address."""

    rssi = property(__get_rssi, __set_rssi)
    """Integer. Received Signal Strength Indicator (RSSI) value."""

    receive_options = property(__get_options, __set_options)
    """Integer. Receive options bitfield."""

    rf_data = property(__get_rf_data, __set_rf_data)
    """Bytearray. Received RF data."""

    io_sample = property(__get_io_sample, __set_io_sample)
    """:class:`.IOSample`: IO sample corresponding to the data contained in the packet."""


class RX16IOPacket(XBeeAPIPacket):
    """
    This class represents an RX16 address IO packet. Packet is built using the
    parameters of the constructor or providing a valid byte array.
    
    I/O data is sent out the UART using an API frame.
    
    .. seealso::
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 14

    def __init__(self, x16bit_addr, rssi, receive_options, rf_data):
        """
        Class constructor. Instantiates an :class:`.RX16IOPacket` object with the provided parameters.
        
        Args:
            x16bit_addr (:class:`.XBee16BitAddress`): the 16-bit source address.
            rssi (Integer): received signal strength indicator.
            receive_options (Integer): bitfield indicating the receive options.
            rf_data (Bytearray): received RF data.
            
        .. seealso::
           | :class:`.ReceiveOptions`
           | :class:`.XBee16BitAddress`
           | :class:`.XBeeAPIPacket`
        """
        super().__init__(ApiFrameType.RX_IO_16)
        self.__x16bit_addr = x16bit_addr
        self.__rssi = rssi
        self.__options = receive_options
        self.__rf_data = rf_data
        self.__io_sample = IOSample(rf_data) if rf_data is not None and len(rf_data) >= 5 else None

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.
        
        Returns:
            :class:`.RX16IOPacket`.
            
        Raises:
            InvalidPacketException: if the bytearray length is less than 14. (start delim. + length (2 bytes) + frame
                type + 16bit addr. + rssi + receive options + rf data (5 bytes) + checksum = 14 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :attr:`.ApiFrameType.RX_IO_16`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.
            
        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=RX16IOPacket.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.RX_IO_16.code:
            raise InvalidPacketException("This packet is not an RX 16 IO packet.")

        return RX16IOPacket(XBee16BitAddress(raw[4:6]), raw[6], raw[7], raw[8:-1])

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return False

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        ret = self.__x16bit_addr.address
        ret.append(self.__rssi)
        ret.append(self.__options)
        if self.__rf_data is not None:
            ret += self.__rf_data
        return ret

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        base = {DictKeys.X16BIT_ADDR:     self.__x16bit_addr.address,
                DictKeys.RSSI:            self.__rssi,
                DictKeys.RECEIVE_OPTIONS: self.__options}

        if self.__io_sample is not None:
            base[DictKeys.NUM_SAMPLES] = 1
            base[DictKeys.DIGITAL_MASK] = self.__io_sample.digital_mask
            base[DictKeys.ANALOG_MASK] = self.__io_sample.analog_mask

            # Digital values
            for i in range(16):
                if self.__io_sample.has_digital_value(IOLine.get(i)):
                    base[IOLine.get(i).description + "digital value"] = \
                        utils.hex_to_string(self.__io_sample.get_digital_value(IOLine.get(i)))

            # Analog values
            for i in range(6):
                if self.__io_sample.has_analog_value(IOLine.get(i)):
                    base[IOLine.get(i).description + "analog value"] = \
                        utils.hex_to_string(self.__io_sample.get_analog_value(IOLine.get(i)))

            # Power supply
            if self.__io_sample.has_power_supply_value():
                base["Power supply value "] = "%02X" % self.__io_sample.power_supply_value

        elif self.__rf_data is not None:
            base[DictKeys.RF_DATA] = utils.hex_to_string(self.__rf_data)

        return base

    def __get_16bit_addr(self):
        """
        Returns the 16-bit source address.

        Returns:
            :class:`.XBee16BitAddress`: the 16-bit source address.

        .. seealso::
           | :class:`.XBee16BitAddress`
        """
        return self.__x16bit_addr

    def __set_16bit_addr(self, x16bit_addr):
        """
        Sets the 16-bit source address.

        Args:
            x16bit_addr (:class:`.XBee16BitAddress`): the new 16-bit source address.

        .. seealso::
           | :class:`.XBee16BitAddress`
        """
        self.__x16bit_addr = x16bit_addr

    def __get_rssi(self):
        """
        Returns the received Signal Strength Indicator (RSSI).

        Returns:
            Integer: the received Signal Strength Indicator (RSSI).
        """
        return self.__rssi

    def __set_rssi(self, rssi):
        """
        Sets the received Signal Strength Indicator (RSSI).

        Args:
            rssi (Integer): the new received Signal Strength Indicator (RSSI).

        """
        self.__rssi = rssi

    def __get_options(self):
        """
        Returns the receive options bitfield.

        Returns:
            Integer: the receive options bitfield.

        .. seealso::
           | :class:`.ReceiveOptions`
        """
        return self.__receive_options

    def __set_options(self, receive_options):
        """
        Sets the receive options bitfield.

        Args:
            receive_options (Integer): the new receive options bitfield.

        .. seealso::
           | :class:`.ReceiveOptions`
        """
        self.__receive_options = receive_options

    def __get_rf_data(self):
        """
        Returns the received RF data.

        Returns:
            Bytearray: the received RF data.
        """
        if self.__rf_data is None:
            return None
        return self.__rf_data.copy()

    def __set_rf_data(self, rf_data):
        """
        Sets the received RF data.

        Args:
            rf_data (Bytearray): the new received RF data.
        """
        if rf_data is None:
            self.__rf_data = None
        else:
            self.__rf_data = rf_data.copy()

        # Modify the ioSample accordingly
        if rf_data is not None and len(rf_data) >= 5:
            self.__io_sample = IOSample(self.__rf_data)
        else:
            self.__io_sample = None

    def __get_io_sample(self):
        """
        Returns the IO sample corresponding to the data contained in the packet.

        Returns:
            :class:`.IOSample`: the IO sample of the packet, ``None`` if the packet has not any data or if the
                sample could not be generated correctly.

        .. seealso::
           | :class:`.IOSample`
        """
        return self.__io_sample

    def __set_io_sample(self, io_sample):
        """
        Sets the IO sample of the packet.

        Args:
            io_sample (:class:`.IOSample`): the new IO sample to set.

        .. seealso::
           | :class:`.IOSample`
        """
        self.__io_sample = io_sample

    x16bit_source_addr = property(__get_16bit_addr, __set_16bit_addr)
    """:class:`.XBee16BitAddress`. 16-bit source address."""

    rssi = property(__get_rssi, __set_rssi)
    """Integer. Received Signal Strength Indicator (RSSI) value."""

    receive_options = property(__get_options, __set_options)
    """Integer. Receive options bitfield."""

    rf_data = property(__get_rf_data, __set_rf_data)
    """Bytearray. Received RF data."""

    io_sample = property(__get_io_sample, __set_io_sample)
    """:class:`.IOSample`: IO sample corresponding to the data contained in the packet."""



# Copyright 2017, Digi International Inc.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from ipaddress import IPv4Address
from digi.xbee.devices import NBIoTDevice
from digi.xbee.models.protocol import IPProtocol

# TODO: Replace with the serial port where your local module is connected to. 
PORT = "COM1"
# TODO: Replace with the baud rate of your local module.
BAUD_RATE = 9600
# TODO: Replace with the destination IP address.
DEST_IP_ADDRESS = "192.168.1.2"
# TODO: Replace with the destination port number (in decimal format).
DEST_PORT = 9750

PROTOCOL = IPProtocol.UDP
DATA_TO_SEND = "Hello XBee!"


def main():
    print(" +------------------------------------------+")
    print(" | XBee Python Library Send UDP Data Sample |")
    print(" +------------------------------------------+\n")

    device = NBIoTDevice(PORT, BAUD_RATE)

    try:
        device.open()

        if not device.is_connected():
            print(">> Error: the device is not connected to the network")
            return

        print("Sending data to %s:%d >> %s..." % (DEST_IP_ADDRESS, DEST_PORT, DATA_TO_SEND))

        device.send_ip_data(IPv4Address(DEST_IP_ADDRESS), DEST_PORT, PROTOCOL, DATA_TO_SEND)

        print("Success")

    finally:
        if device is not None and device.is_open():
            device.close()


if __name__ == '__main__':
    main()



# Copyright 2017, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging


# Number of bits to extract with the mask (__MASK)
__MASK_NUM_BITS = 8

# Bit mask to extract the less important __MAS_NUM_BITS bits of a number.
__MASK = 0xFF


def is_bit_enabled(number, position):
    """
    Returns whether the bit located at ``position`` within ``number`` is enabled or not.

    Args:
        number (Integer): the number to check if a bit is enabled.
        position (Integer): the position of the bit to check if is enabled in ``number``.

    Returns:
        Boolean: ``True`` if the bit located at ``position`` within ``number`` is enabled, ``False`` otherwise.
    """
    return ((number & 0xFFFFFFFF) >> position) & 0x01 == 0x01


def hex_string_to_bytes(hex_string):
    """
    Converts a String (composed by hex. digits) into a bytearray with same digits.
    
    Args:
        hex_string (String): String (made by hex. digits) with "0x" header or not.

    Returns:
        Bytearray: bytearray containing the numeric value of the hexadecimal digits.
        
    Raises:
        ValueError: if invalid literal for int() with base 16 is provided.
    
    Example:
        >>> a = "0xFFFE"
        >>> for i in hex_string_to_bytes(a): print(i)
        255
        254
        >>> print(type(hex_string_to_bytes(a)))
        <type 'bytearray'>
        
        >>> b = "FFFE"
        >>> for i in hex_string_to_bytes(b): print(i)
        255
        254
        >>> print(type(hex_string_to_bytes(b)))
        <type 'bytearray'>
        
    """
    aux = int(hex_string, 16)
    return int_to_bytes(aux)


def int_to_bytes(number, num_bytes=None):
    """
    Converts the provided integer into a bytearray.
    
    If ``number`` has less bytes than ``num_bytes``, the resultant bytearray
    is filled with zeros (0x00) starting at the beginning.
    
    If ``number`` has more bytes than ``num_bytes``, the resultant bytearray
    is returned without changes.
    
    Args:
        number (Integer): the number to convert to a bytearray.
        num_bytes (Integer): the number of bytes that the resultant bytearray will have.

    Returns:
        Bytearray: the bytearray corresponding to the provided number.

    Example:
        >>> a=0xFFFE
        >>> print([i for i in int_to_bytes(a)])
        [255,254]
        >>> print(type(int_to_bytes(a)))
        <type 'bytearray'>
        
    """
    byte_array = bytearray()
    byte_array.insert(0, number & __MASK)
    number >>= __MASK_NUM_BITS
    while number != 0:
        byte_array.insert(0, number & __MASK)
        number >>= __MASK_NUM_BITS

    if num_bytes is not None:
        while len(byte_array) < num_bytes:
            byte_array.insert(0, 0x00)

    return byte_array


def length_to_int(byte_array):
    """
    Calculates the length value for the given length field of a packet.
    Length field are bytes 1 and 2 of any packet.
    
    Args:
        byte_array (Bytearray): length field of a packet.
        
    Returns:
        Integer: the length value.
    
    Raises:
        ValueError: if ``byte_array`` is not a valid length field (it has length distinct than 0).
    
    Example:
        >>> b = bytearray([13,14])
        >>> c = length_to_int(b)
        >>> print("0x%02X" % c)
        0x1314
        >>> print(c)
        4884
    """
    if len(byte_array) != 2:
        raise ValueError("bArray must have length 2")
    return (byte_array[0] << 8) + byte_array[1]


def bytes_to_int(byte_array):
    """
    Converts the provided bytearray in an Integer.
    This integer is result of concatenate all components of ``byte_array``
    and convert that hex number to a decimal number.

    Args:
        byte_array (Bytearray): bytearray to convert in integer.

    Returns:
        Integer: the integer corresponding to the provided bytearray.

    Example:
        >>> x = bytearray([0xA,0x0A,0x0A]) #this is 0xA0A0A
        >>> print(bytes_to_int(x))
        657930
        >>> b = bytearray([0x0A,0xAA])    #this is 0xAAA
        >>> print(bytes_to_int(b))
        2730
    """
    if len(byte_array) == 0:
        return 0
    return int("".join(["%02X" % i for i in byte_array]), 16)


def ascii_to_int(ni):
    """
    Converts a bytearray containing the ASCII code of each number digit in an Integer.
    This integer is result of the number formed by all ASCII codes of the bytearray.
    
    Example:
        >>> x = bytearray( [0x31,0x30,0x30] )   #0x31 => ASCII code for number 1.
                                                #0x31,0x30,0x30 <==> 1,0,0
        >>> print(ascii_to_int(x))
        100
    """
    return int("".join([str(i - 0x30) for i in ni]))


def int_to_ascii(number):
    """
    Converts an integer number to a bytearray. Each element of the bytearray is the ASCII
    code that corresponds to the digit of its position.

    Args:
        number (Integer): the number to convert to an ASCII bytearray.

    Returns:
        Bytearray: the bytearray containing the ASCII value of each digit of the number.

    Example:
        >>> x = int_to_ascii(100)
        >>> print(x)
        100
        >>> print([i for i in x])
        [49, 48, 48]
    """
    return bytearray([ord(i) for i in str(number)])


def int_to_length(number):
    """
    Converts am integer into a bytearray of 2 bytes corresponding to the length field of a
    packet. If this bytearray has length 1, a byte with value 0 is added at the beginning.

    Args:
        number (Integer): the number to convert to a length field.

    Returns:


    Raises:
        ValueError: if ``number`` is less than 0 or greater than 0xFFFF.
        
    Example:
        >>> a = 0
        >>> print(hex_to_string(int_to_length(a)))
        00 00
        
        >>> a = 8
        >>> print(hex_to_string(int_to_length(a)))
        00 08
        
        >>> a = 200
        >>> print(hex_to_string(int_to_length(a)))
        00 C8
        
        >>> a = 0xFF00
        >>> print(hex_to_string(int_to_length(a)))
        FF 00
        
        >>> a = 0xFF
        >>> print(hex_to_string(int_to_length(a)))
        00 FF
    """
    if number < 0 or number > 0xFFFF:
        raise ValueError("The number must be between 0 and 0xFFFF.")
    length = int_to_bytes(number)
    if len(length) < 2:
        length.insert(0, 0)
    return length


def hex_to_string(byte_array):
    """
    Returns the provided bytearray in a pretty string format. All bytes are separated by blank spaces and
    printed in hex format.

    Args:
        byte_array (Bytearray): the bytearray to print in pretty string.

    Returns:
        String: the bytearray formatted in a pretty string.
    """
    return " ".join(["%02X" % i for i in byte_array])


def doc_enum(enum_class, descriptions=None):
    """
    Returns a string with the description of each value of an enumeration.
    
    Args:
        enum_class (Enumeration): the Enumeration to get its values documentation.
        descriptions (dictionary): each enumeration's item description. The key is the enumeration element name
            and the value is the description.
            
    Returns:
        String: the string listing all the enumeration values and their descriptions.
    """
    tab = " "*4
    data = "\n| Values:\n"
    for x in enum_class:
        data += """| {:s}**{:s}**{:s} {:s}\n""".format(tab, x,
                                                       ":" if descriptions is not None else " =",
                                                       str(x.value) if descriptions is None else descriptions[x])
    return data + "| \n"


def enable_logger(name, level=logging.DEBUG):
    """
    Enables a logger with the given name and level.

    Args:
        name (String): name of the logger to enable.
        level (Integer): logging level value.
    
    Assigns a default formatter and a default handler (for console).
    """
    log = logging.getLogger(name)
    log.disabled = False
    ch = logging.StreamHandler()
    ch.setLevel(level)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)-7s - %(message)s')
    ch.setFormatter(formatter)
    log.addHandler(ch)
    log.setLevel(level)


def disable_logger(name):
    """
    Disables the logger with the give name.

    Args:
        name (String): the name of the logger to disable.
    """
    log = logging.getLogger(name)
    log.disabled = True



# Copyright 2017, 2018, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from digi.xbee.util import utils
from enum import Enum, unique
from digi.xbee.exception import OperationNotSupportedException


@unique
class IOLine(Enum):
    """
    Enumerates the different IO lines that can be found in the XBee devices. 

    Depending on the hardware and firmware of the device, the number of lines 
    that can be used as well as their functionality may vary. Refer to the 
    product manual to learn more about the IO lines of your XBee device.
    """

    DIO0_AD0 = ("DIO0/AD0", 0, "D0")
    DIO1_AD1 = ("DIO1/AD1", 1, "D1")
    DIO2_AD2 = ("DIO2/AD2", 2, "D2")
    DIO3_AD3 = ("DIO3/AD3", 3, "D3")
    DIO4_AD4 = ("DIO4/AD4", 4, "D4")
    DIO5_AD5 = ("DIO5/AD5", 5, "D5")
    DIO6 = ("DIO6", 6, "D6")
    DIO7 = ("DIO7", 7, "D7")
    DIO8 = ("DIO8", 8, "D8")
    DIO9 = ("DIO9", 9, "D9")
    DIO10_PWM0 = ("DIO10/PWM0", 10, "P0", "M0")
    DIO11_PWM1 = ("DIO11/PWM1", 11, "P1", "M1")
    DIO12 = ("DIO12", 12, "P2")
    DIO13 = ("DIO13", 13, "P3")
    DIO14 = ("DIO14", 14, "P4")
    DIO15 = ("DIO15", 15, "P5")
    DIO16 = ("DIO16", 16, "P6")
    DIO17 = ("DIO17", 17, "P7")
    DIO18 = ("DIO18", 18, "P8")
    DIO19 = ("DIO19", 19, "P9")

    def __init__(self, description, index, at_command, pwm_command=None):
        self.__description = description
        self.__index = index
        self.__at_command = at_command
        self.__pwm_command = pwm_command

    def __get_description(self):
        """
        Returns the description of the IOLine element.

        Returns:
            String: the description of the IOLine element.
        """
        return self.__description

    def __get_index(self):
        """
        Returns the index of the IOLine element.

        Returns:
            Integer: the index of the IOLine element.
        """
        return self.__index

    def __get_at_command(self):
        """
        Returns the AT command of the IOLine element.

        Returns:
            String: the AT command of the IOLine element.
        """
        return self.__at_command

    def __get_pwm_command(self):
        """
        Returns the PWM AT command associated to the IOLine element.

        Returns:
            String: the PWM AT command associated to the IO line, ``None`` if the IO line does not have a PWM
                AT command associated.
        """
        return self.__pwm_command

    def has_pwm_capability(self):
        """
        Returns whether the IO line has PWM capability or not.

        Returns:
            Boolean: ``True`` if the IO line has PWM capability, ``False`` otherwise.
        """
        return self.__pwm_command is not None

    @classmethod
    def get(cls, index):
        """
        Returns the :class:`.IOLine` for the given index.

        Args:
            index (Integer): Returns the :class:`.IOLine` for the given index.

        Returns:
            :class:`.IOLine`: :class:`.IOLine` with the given code, ``None`` if there is not any line with that index.
        """
        try:
            return cls.lookupTable[index]
        except KeyError:
            return None

    description = property(__get_description)
    """String. The IO line description."""

    index = property(__get_index)
    """Integer. The IO line index."""

    at_command = property(__get_at_command)
    """String. The IO line AT command."""

    pwm_at_command = property(__get_pwm_command)
    """String. The IO line PWM AT command."""


IOLine.lookupTable = {x.index: x for x in IOLine}
IOLine.__doc__ += utils.doc_enum(IOLine)


@unique
class IOValue(Enum):
    """
    Enumerates the possible values of a :class:`.IOLine` configured as digital I/O.
    """

    LOW = 4
    HIGH = 5

    def __init__(self, code):
        self.__code = code

    def __get_code(self):
        """
        Returns the code of the IOValue element.

        Returns:
            String: the code of the IOValue element.
        """
        return self.__code

    @classmethod
    def get(cls, code):
        """
        Returns the IOValue for the given code.

        Args:
            code (Integer): the code corresponding to the IOValue to get.

        Returns:
            :class:`.IOValue`: the IOValue with the given code, ``None`` if there is not any IOValue with that code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return None

    code = property(__get_code)
    """Integer. The IO value code."""


IOValue.lookupTable = {x.code: x for x in IOValue}


class IOSample(object):
    """
    This class represents an IO Data Sample. The sample is built using the
    the constructor. The sample contains an analog and digital mask indicating 
    which IO lines are configured with that functionality.

    Depending on the protocol the XBee device is executing, the digital and 
    analog masks are retrieved in separated bytes (2 bytes for the digital mask 
    and 1 for the analog mask) or merged contained (digital and analog masks 
    are contained in 2 bytes). 

    Digital and analog channels masks
    Indicates which digital and ADC IO lines are configured in the module. Each
    bit corresponds to one digital or ADC IO line on the module:
    ::

            bit 0 =  DIO01
            bit 1 =  DIO10
            bit 2 =  DIO20
            bit 3 =  DIO31
            bit 4 =  DIO40
            bit 5 =  DIO51
            bit 6 =  DIO60
            bit 7 =  DIO70
            bit 8 =  DIO80
            bit 9 =  AD00
            bit 10 = AD11
            bit 11 = AD21
            bit 12 = AD30
            bit 13 = AD40
            bit 14 = AD50
            bit 15 = NA0

            Example: mask of 0x0C29 means DIO0, DIO3, DIO5, AD1 and AD2 enabled.
            0 0 0 0 1 1 0 0 0 0 1 0 1 0 0 1

    Digital Channel Mask
    Indicates which digital IO lines are configured in the module. Each bit 
    corresponds to one digital IO line on the module:
    ::

            bit 0 =  DIO0AD0
            bit 1 =  DIO1AD1 
            bit 2 =  DIO2AD2
            bit 3 =  DIO3AD3
            bit 4 =  DIO4AD4
            bit 5 =  DIO5AD5ASSOC
            bit 6 =  DIO6RTS
            bit 7 =  DIO7CTS
            bit 8 =  DIO8DTRSLEEP_RQ
            bit 9 =  DIO9ON_SLEEP
            bit 10 = DIO10PWM0RSSI
            bit 11 = DIO11PWM1
            bit 12 = DIO12CD
            bit 13 = DIO13
            bit 14 = DIO14
            bit 15 = NA

            Example: mask of 0x040B means DIO0, DIO1, DIO2, DIO3 and DIO10 enabled.
            0 0 0 0 0 1 0 0 0 0 0 0 1 0 1 1

    Analog Channel Mask
    Indicates which lines are configured as ADC. Each bit in the analog 
    channel mask corresponds to one ADC line on the module.
    ::

            bit 0 = AD0DIO0
            bit 1 = AD1DIO1
            bit 2 = AD2DIO2
            bit 3 = AD3DIO3
            bit 4 = AD4DIO4
            bit 5 = AD5DIO5ASSOC
            bit 6 = NA
            bit 7 = Supply Voltage Value

            Example: mask of 0x83 means AD0, and AD1 enabled.
            0 0 0 0 0 0 1 1
    """

    __pattern = "[{key}: {value}], "
    """Pattern for digital and analog values in __str__ method."""

    __pattern2 = "[Power supply voltage: {value}], "
    """Pattern for power supply voltage in __str__ method."""

    __MIN_IO_SAMPLE_PAYLOAD_LENGTH = 5

    def __init__(self, io_sample_payload):
        """
        Class constructor. Instantiates a new :class:`.IOSample` object with the provided parameters.

        Args:
            io_sample_payload (Bytearray): The payload corresponding to an IO sample.

        Raises:
            ValueError: if io_sample_payload length is less than 5.
        """
        # dictionaries
        self.__digital_values_map = {}  # {IOLine : IOValue}
        self.__analog_values_map = {}  # {IOLine : Integer}

        # Integers:
        self.__digital_hsb_mask = None
        self.__digital_lsb_mask = None
        self.__digital_mask = None
        self.__analog_mask = None
        self.__digital_hsb_values = None
        self.__digital_lsb_values = None
        self.__digital_values = None
        self.__power_supply_voltage = None

        if len(io_sample_payload) < IOSample.__MIN_IO_SAMPLE_PAYLOAD_LENGTH:
            raise ValueError("IO sample payload must be longer than 4.")

        self.__io_sample_payload = io_sample_payload

        if len(self.__io_sample_payload) % 2 != 0:
            self.__parse_raw_io_sample()
        else:
            self.__parse_io_sample()

    def __str__(self):
        s = "{"
        if self.has_digital_values():
            s += (''.join([self.__pattern.format(key=x, value=self.__digital_values_map[x]) for x in
                           self.__digital_values_map.keys()]))
        if self.has_analog_values():
            s += (''.join([self.__pattern.format(key=x, value=self.__analog_values_map[x]) for x in
                           self.__analog_values_map.keys()]))
        if self.has_power_supply_value():
            try:
                s += self.__pattern2.format(value=self.__power_supply_voltage)
            except OperationNotSupportedException:
                pass
        s += "}"
        aux = s.replace(", }", "}")
        return aux

    @staticmethod
    def min_io_sample_payload():
        """
        Returns  the minimum IO sample payload length.

        Returns:
            Integer: the minimum IO sample payload length.
        """
        return IOSample.__MIN_IO_SAMPLE_PAYLOAD_LENGTH

    def __parse_raw_io_sample(self):
        """
        Parses the information contained in the IO sample bytes reading the 
        value of each configured DIO and ADC.
        """
        data_index = 3

        # Obtain the digital mask.                                    # Available digital IOs in 802.15.4
        self.__digital_hsb_mask = self.__io_sample_payload[1] & 0x01  # 0 0 0 0 0 0 0 1
        self.__digital_lsb_mask = self.__io_sample_payload[2] & 0xFF  # 1 1 1 1 1 1 1 1
        # Combine the masks.
        self.__digital_mask = (self.__digital_hsb_mask << 8) + self.__digital_lsb_mask
        # Obtain the analog mask.
        self.__analog_mask = ((self.__io_sample_payload[1] << 8)       # Available analog IOs in 802.15.4
                              + self.__io_sample_payload[2]) & 0x7E00  # 0 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0

        # Read the digital values (if any). There are 9 possible digital lines in
        # 802.15.4 protocol. The digital mask indicates if there is any digital
        # line enabled to read its value. If 0, no digital values are received.
        if self.__digital_mask > 0:
            # Obtain the digital values.
            self.__digital_hsb_values = self.__io_sample_payload[3] & 0x7F
            self.__digital_lsb_values = self.__io_sample_payload[4] & 0xFF
            # Combine the values.
            self.__digital_values = (self.__digital_hsb_values << 8) + self.__digital_lsb_values

            for i in range(16):
                if not utils.is_bit_enabled(self.__digital_mask, i):
                    continue
                if utils.is_bit_enabled(self.__digital_values, i):
                    self.__digital_values_map[IOLine.get(i)] = IOValue.HIGH
                else:
                    self.__digital_values_map[IOLine.get(i)] = IOValue.LOW

            # Increase the data index to read the analog values.
            data_index += 2

        # Read the analog values (if any). There are 6 possible analog lines.
        # The analog mask indicates if there is any analog line enabled to read
        # its value. If 0, no analog values are received.
        adc_index = 9
        while (len(self.__io_sample_payload) - data_index) > 1 and adc_index < 16:
            if not (utils.is_bit_enabled(self.__analog_mask, adc_index)):
                adc_index += 1
                continue

            # 802.15.4 protocol does not provide power supply value, so get just the ADC data.
            self.__analog_values_map[IOLine.get(adc_index - 9)] = \
                ((self.__io_sample_payload[data_index] & 0xFF) << 8) + (self.__io_sample_payload[data_index + 1] & 0xFF)
            # Increase the data index to read the next analog values.
            data_index += 2
            adc_index += 1

    def __parse_io_sample(self):
        """
        Parses the information contained in the IO sample bytes reading the 
        value of each configured DIO and ADC.
        """
        data_index = 4

        # Obtain the digital masks.                                   # Available digital IOs
        self.__digital_hsb_mask = self.__io_sample_payload[1] & 0x7F  # 0 1 1 1 1 1 1 1
        self.__digital_lsb_mask = self.__io_sample_payload[2] & 0xFF  # 1 1 1 1 1 1 1 1
        # Combine the masks.
        self.__digital_mask = (self.__digital_hsb_mask << 8) + self.__digital_lsb_mask
        # Obtain the analog mask.                                # Available analog IOs
        self.__analog_mask = self.__io_sample_payload[3] & 0xBF  # 1 0 1 1 1 1 1 1

        # Read the digital values (if any). There are 16 possible digital lines.
        # The digital mask indicates if there is any digital line enabled to read
        # its value. If 0, no digital values are received.
        if self.__digital_mask > 0:
            # Obtain the digital values.
            self.__digital_hsb_values = self.__io_sample_payload[4] & 0x7F
            self.__digital_lsb_values = self.__io_sample_payload[5] & 0xFF
            # Combine the values.
            self.__digital_values = (self.__digital_hsb_values << 8) + self.__digital_lsb_values

            for i in range(16):
                if not utils.is_bit_enabled(self.__digital_mask, i):
                    continue
                if utils.is_bit_enabled(self.__digital_values, i):
                    self.__digital_values_map[IOLine.get(i)] = IOValue.HIGH
                else:
                    self.__digital_values_map[IOLine.get(i)] = IOValue.LOW
            # Increase the data index to read the analog values.
            data_index += 2

        # Read the analog values (if any). There are 6 possible analog lines.
        # The analog mask indicates if there is any analog line enabled to read
        # its value. If 0, no analog values are received.
        adc_index = 0
        while (len(self.__io_sample_payload) - data_index) > 1 and adc_index < 8:
            if not utils.is_bit_enabled(self.__analog_mask, adc_index):
                adc_index += 1
                continue
            # When analog index is 7, it means that the analog value corresponds to the power
            # supply voltage, therefore this value should be stored in a different value.
            if adc_index == 7:
                self.__power_supply_voltage = ((self.__io_sample_payload[data_index] & 0xFF) << 8) + \
                                              (self.__io_sample_payload[data_index + 1] & 0xFF)
            else:
                self.__analog_values_map[IOLine.get(adc_index)] = \
                    ((self.__io_sample_payload[data_index] & 0xFF) << 8) + \
                    (self.__io_sample_payload[data_index + 1] & 0xFF)
            # Increase the data index to read the next analog values.
            data_index += 2
            adc_index += 1

    def __get_digital_hsb_mask(self):
        """
        Returns the High Significant Byte (HSB) of the digital mask.

        Returns:
            Integer: the HSB of the digital mask.
        """
        return self.__digital_hsb_mask

    def __get_digital_lsb_mask(self):
        """
        Returns the Low Significant Byte (HSB) of the digital mask.

        Returns:
            Integer: the LSB of the digital mask.
        """
        return self.__digital_lsb_mask

    def __get_digital_mask(self):
        """
        Returns the combined (HSB + LSB) of the digital mask.

        Returns:
            Integer: the digital mask.
        """
        return self.__digital_mask

    def __get_digital_values(self):
        """
        Returns the digital values map.

        To verify if this sample contains a valid digital values, use the 
        method :meth:`.IOSample.has_digital_values`.

        Returns:
            Dictionary: the digital values map.
        """
        return self.__digital_values_map.copy()

    def __get_analog_mask(self):
        """
        Returns the analog mas.

        Returns:
            Integer: the analog mask.
        """
        return self.__analog_mask

    def __get_analog_values(self):
        """
        Returns the analog values map.

        To verify if this sample contains a valid analog values, use the 
        method :meth:`.IOSample.has_analog_values`.

        Returns:
            Dictionary: the analog values map.
        """
        return self.__analog_values_map.copy()

    def __get_power_supply_value(self):
        """
        Returns the value of the power supply voltage.

        To verify if this sample contains the power supply voltage, use the 
        method :meth:`.IOSample.has_power_supply_value`.

        Returns:
            Integer: the power supply value, ``None`` if the sample does not contain power supply value.
        """
        return self.__power_supply_voltage if self.has_power_supply_value() else None

    def has_digital_values(self):
        """
        Checks whether the IOSample has digital values or not.

        Returns:
            Boolean: ``True`` if the sample has digital values, ``False`` otherwise.
        """
        return len(self.__digital_values_map) > 0

    def has_digital_value(self, io_line):
        """
        Returns whether th IO sample contains a digital value for the provided IO line or not.

        Args:
            io_line (:class:`IOLine`): The IO line to check if it has a digital value.

        Returns:
            Boolean: ``True`` if the given IO line has a digital value, ``False`` otherwise.
        """
        return io_line in self.__digital_values_map.keys()

    def has_analog_value(self, io_line):
        """
        Returns whether the given IOLine has an analog value or not.

        Returns:
            Boolean: ``True`` if the given IOLine has an analog value, ``False`` otherwise.
        """
        return io_line in self.__analog_values_map.keys()

    def has_analog_values(self):
        """
        Returns whether the {@code IOSample} has analog values or not.

        Returns:
            Boolean. ``True`` if there are analog values, ``False`` otherwise.
        """
        return len(self.__analog_values_map) > 0

    def has_power_supply_value(self):
        """
        Returns whether the IOSample has power supply value or not.

        Returns:
            Boolean. ``True`` if the given IOLine has a power supply value, ``False`` otherwise.
        """
        return utils.is_bit_enabled(self.__analog_mask, 7)

    def get_digital_value(self, io_line):
        """
        Returns the digital value of the provided IO line.

        To verify if this sample contains a digital value for the given :class:`.IOLine`,
        use the method :meth:`.IOSample.has_digital_value`.

        Args:
            io_line (:class:`.IOLine`): The IO line to get its digital value.

        Returns:
            :class:`.IOValue`: The :class:`.IOValue` of the given IO line or ``None`` if the
                IO sample does not contain a digital value for the given IO line.

        .. seealso::
           | :class:`.IOLine`
           | :class:`.IOValue`
        """
        if io_line in self.__digital_values_map:
            return self.__digital_values_map[io_line]
        return None

    def get_analog_value(self, io_line):
        """
        Returns the analog value of the provided IO line.

        To verify if this sample contains an analog value for the given :class:`.IOLine`,
        use the method :meth:`.IOSample.has_analog_value`.

        Args:
            io_line (:class:`.IOLine`): The IO line to get its analog value.

        Returns:
            Integer: The analog value of the given IO line or ``None`` if the IO sample does not
                contain an analog value for the given IO line.

        .. seealso::
           | :class:`.IOLine`
        """
        if io_line in self.__analog_values_map:
            return self.__analog_values_map[io_line]
        return None

    digital_hsb_mask = property(__get_digital_hsb_mask)
    """Integer. High Significant Byte (HSB) of the digital mask."""

    digital_lsb_mask = property(__get_digital_lsb_mask)
    """Integer. Low Significant Byte (LSB) of the digital mask."""

    digital_mask = property(__get_digital_mask)
    """Integer. Digital mask of the IO sample."""

    analog_mask = property(__get_analog_mask)
    """Integer. Analog mask of the IO sample."""

    digital_values = property(__get_digital_values)
    """Dictionary. Digital values map."""

    analog_values = property(__get_analog_values)
    """Dictionary. Analog values map."""

    power_supply_value = property(__get_power_supply_value)
    """Integer. Power supply value, ``None`` if the sample does not contain power supply value."""


class IOMode(Enum):
    """
    Enumerates the different Input/Output modes that an IO line can be 
    configured with.
    """

    DISABLED = 0
    """Disabled"""

    SPECIAL_FUNCTIONALITY = 1
    """Firmware special functionality"""

    PWM = 2
    """PWM output"""

    ADC = 2
    """Analog to Digital Converter"""

    DIGITAL_IN = 3
    """Digital input"""

    DIGITAL_OUT_LOW = 4
    """Digital output, Low"""

    DIGITAL_OUT_HIGH = 5
    """Digital output, High"""



# Copyright 2017, Digi International Inc.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from digi.xbee.devices import CellularDevice

# TODO: Replace with the serial port where your local module is connected to. 
PORT = "COM1"
# TODO: Replace with the baud rate of your local module.
BAUD_RATE = 9600
# TODO: Replace with the phone number of the device to send the SMS to.
PHONE = ""
# TODO: Optionally, replace with the text of the SMS.
SMS_TEXT = "Hello from XBee Cellular!"


def main():
    print(" +-------------------------------------+")
    print(" | XBee Python Library Send SMS Sample |")
    print(" +-------------------------------------+\n")

    device = CellularDevice(PORT, BAUD_RATE)

    try:
        device.open()

        print("Sending SMS to %s >> %s..." % (PHONE, SMS_TEXT))

        device.send_sms(PHONE, SMS_TEXT)

        print("Success")

    finally:
        if device is not None and device.is_open():
            device.close()


if __name__ == '__main__':
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2017, 2018, Digi International Inc. All Rights Reserved.

from setuptools import setup, find_packages
from codecs import open
from os import path


DEPENDENCIES = (
    'pyserial>=3',
    'ipaddress>=1',
    'enum34>=1',
    'futures>=3'
)

here = path.abspath(path.dirname(__file__))
 
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()
 
setup(
    name='ngcp-xbee',
    version='1.3',
    description='Digi XBee Python library, compatible with Python 2',
    long_description=long_description,
    url='https://github.com/tianatonnu/python-xbee.git',
    author='Tiana Ton Nu, Digi International Inc.',
    author_email='tianattonnu@gmail.com, tech.support@digi.com',
    packages=find_packages(exclude=('unit_test*', 'functional_tests*', 'demos*')),
    keywords=['xbee', 'IOT', 'wireless', 'radio frequency'],
    license='Mozilla Public License 2.0 (MPL 2.0)',
    python_requires='>=2',
    install_requires=[
        'pyserial>=3',
        'ipaddress>=1',
        'enum34>=1',
        'futures>=3'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Telecommunications Industry',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Software Development :: Libraries',
        'Topic :: Home Automation',
        'Topic :: Games/Entertainment',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Python :: 2',
        'Operating System :: OS Independent',
    ],
)
